/*******************************************************************************
 * Copyright (c) 2009 Ericsson
 * 
 * All rights reserved. This program and the accompanying materials are
 * made available under the terms of the Eclipse Public License v1.0 which
 * accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *   Francois Chouinard - Initial API and implementation
 *******************************************************************************/

package org.eclipse.linuxtools.tmf.trace;

import org.eclipse.linuxtools.tmf.event.TmfEvent;
import org.eclipse.linuxtools.tmf.event.TmfTimeRange;
import org.eclipse.linuxtools.tmf.event.TmfTimestamp;

/**
 * <b><u>ITmfTrace</u></b>
 * <p>
 */
public interface ITmfTrace {

	/**
	 * @return the trace path 
	 */
	public String getPath();
    
	/**
	 * @return the trace name 
	 */
	public String getName();

	/**
	 * @return the number of events in the trace
	 */
	public long getNbEvents();
    
	/**
	 * Trace time range accessors
	 */
	public TmfTimeRange getTimeRange();
    public TmfTimestamp getStartTime();
    public TmfTimestamp getEndTime();

    /**
     * Positions the trace at the first event with the specified
     * timestamp or index (i.e. the nth event in the trace).
     * 
     * Returns a context which can later be used to read the event.
     * 
     * @param data.timestamp
     * @param data.index
     * @return a context object for subsequent reads
     */
    public TmfTraceContext seekLocation(Object location);
    public TmfTraceContext seekEvent(TmfTimestamp timestamp);
    public TmfTraceContext seekEvent(long index);

    /**
     * Return the event pointed by the supplied context (or null if
     * no event left) and updates the context to the next event.
     * 
     * @return the next event in the stream
     */
    public TmfEvent getNextEvent(TmfTraceContext context);

	/**
	 * Skips the given number of events or until the given time stamp is
	 * reached,
	 * updates the context to the next event, and returns the number of events
	 * that have been skipped.
	 * If filter is not null, processing stops when the filter condition is
	 * fulfilled.
	 * 
	 * @param context
	 *            the event context to modify
	 * @param number
	 *            the number of events to skip
	 * @param time
	 *            time stamp to stop
	 * @param filter
	 *            a filter object used to prematurely stop skipping
	 * @param skipped
	 *            an optional array of time stamps, that receive the time stamps
	 *            of skipped events. The maximum number of time stamps set is
	 *            the number of skipped events, as returned by this method. If
	 *            the array is smaller, the time stamps of the first events are
	 *            set
	 * @return the number of skipped events or -1 if no event skipped because
	 *         end-of-trace reached
	 */
    public long skipEvents(TmfTraceContext context, long number, TmfTimestamp time, Object filter, TmfTimestamp[] skipped);

    // ========================================================================
    // Filter and search support
    // ========================================================================

	static class FindResult
	{
		long index;
		TmfEvent event;
	};
	
	/**
	 * Find the next event that matches the filter condition. Returns the event and the index of that event in this trace file.
	 * The context is incremented after the result match - but not necessarily to the next match!
	 * @param context
	 * @param filter
	 * @return the result match, or null if no match found
	 */
	public FindResult findNextEvent( TmfTraceContext context, Object filter );

	/**
	 * Test if the event matches the filter condition. 
	 * @param event
	 * @param filter
	 * @return true if the event matches the filter condition, false otherwise
	 */
	public boolean testEvent( TmfEvent event, Object filter );

}
