/*******************************************************************************
 * Copyright (c) 2009 Ericsson
 * 
 * All rights reserved. This program and the accompanying materials are
 * made available under the terms of the Eclipse Public License v1.0 which
 * accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *   Francois Chouinard - Initial API and implementation
 *   Thomas Gatterweh	- Updated experiment indexing
 *******************************************************************************/

package org.eclipse.linuxtools.tmf.trace;

import java.lang.reflect.InvocationTargetException;
import java.util.Collections;
import java.util.Comparator;
import java.util.TreeSet;
import java.util.Vector;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.linuxtools.tmf.component.TmfComponent;
import org.eclipse.linuxtools.tmf.event.TmfEvent;
import org.eclipse.linuxtools.tmf.event.TmfTimeRange;
import org.eclipse.linuxtools.tmf.event.TmfTimestamp;
import org.eclipse.linuxtools.tmf.request.ITmfRequestHandler;
import org.eclipse.linuxtools.tmf.request.TmfDataRequest;
import org.eclipse.linuxtools.tmf.request.TmfFilteringDataRequest;
import org.eclipse.linuxtools.tmf.signal.TmfSignalHandler;
import org.eclipse.linuxtools.tmf.signal.TmfSignalManager;

/**
 * <b><u>TmfExperiment</u></b>
 * <p>
 * TmfExperiment presents a time-ordered, unified view of a set of TmfTraces
 * that are part of a tracing experiment. 
 * <p>
 */
public class TmfExperiment extends TmfComponent implements ITmfRequestHandler<TmfEvent> {

// TODO: Complete multi-trace experiment
// TODO: Add support for dynamic addition/removal of traces
// TODO: Add support for live streaming (notifications, incremental indexing, ...)

	// ========================================================================
    // Attributes
    // ========================================================================

	// The currently selected experiment
//    private static TmfExperiment fCurrentExperiment;

	// The experiment ID
    protected String fExperimentId;

    // The set of trace sthat constitute the experiment
    protected Vector<ITmfTrace> fTraces;

    // The total number of events
    protected int fNbEvents;

    // The experiment time range
    protected TmfTimeRange fTimeRange;

    // The experiment reference timestamp (default: BigBang)
    protected TmfTimestamp fEpoch;

    // Indicates if the stream should be indexed synchronously (default: false)
    protected final boolean fWaitForIndexCompletion;

    // ========================================================================
    // Constructors/Destructor
    // ========================================================================

    /**
     * @param id
     * @param traces
     */
    public TmfExperiment(String id, ITmfTrace[] traces) {
        this(id, traces, TmfTimestamp.BigBang, DEFAULT_INDEX_PAGE_SIZE, false);
    }

    /**
     * @param id
     * @param traces
     * @param waitForIndexCompletion
     */
    public TmfExperiment(String id, ITmfTrace[] traces, boolean waitForIndexCompletion) {
        this(id, traces, TmfTimestamp.BigBang, DEFAULT_INDEX_PAGE_SIZE, waitForIndexCompletion);
    }

    /**
     * @param id
     * @param traces
     * @param epoch
     * @param waitForIndexCompletion
     */
    public TmfExperiment(String id, ITmfTrace[] traces, TmfTimestamp epoch, int indexPageSize, boolean waitForIndexCompletion) {
    	super();

    	TmfSignalManager.addListener( this );
    	
    	fExperimentId = id;
    	fTraces = new Vector<ITmfTrace>();
    	for (ITmfTrace trace : traces) {
    		fTraces.add(trace);
    	}
    	fEpoch = epoch;
    	fIndexPageSize = indexPageSize;
        fWaitForIndexCompletion = waitForIndexCompletion;

		updateNbEvents();
		updateTimeRange();
//		indexExperiment();
    }

    /**
     * 
     */
    @Override
	public void dispose() {
    	TmfSignalManager.removeListener( this );
 		synchronized( fIndexing )
		{
			if( fIndexing )
			{
		        Job j = job;
		        if( j != null )
		        {
					try
					{
		            	j.cancel();
						fIndexing.wait();
						j.join();
					}
					catch( InterruptedException e )
					{
					}
		        }
			}
		}
    	fTraces.clear();
    	super.dispose();
//    	fCurrentExperiment= null;
    }

    // ========================================================================
    // Accessors
    // ========================================================================

//    public static TmfExperiment getCurrentExperiment() {
//    	return fCurrentExperiment;
//    }

    public String getExperimentId() {
    	return fExperimentId;
    }

    public ITmfTrace[] getTraces() {
    	ITmfTrace[] result = new ITmfTrace[fTraces.size()];
    	return fTraces.toArray(result);
    }

    public TmfTimestamp getEpoch() {
    	return fEpoch;
    }

    public TmfTimeRange getTimeRange() {
    	return fTimeRange;
    }

    public int getNbEvents() {
    	return fNbEvents;
    }

    /**
     * Returns the index of the first event with the requested timestamp.
     * If none, returns the index of the next event (if any).
     *  
     * @param ts
     * @return
     */
    public long getIndex(TmfTimestamp ts) {
    	long index = 0;
    	for( ITmfTrace trace : fTraces )
    	{
    		TmfTraceContext context = trace.seekEvent(ts);
    		index += context.getIndex();
    	}
		return index;
    }

    /**
     * Returns the index of the first event with the requested timestamp.
     * If none, returns the index of the next event (if any).
     * The timestamp of the event at the returned index is put into array
     * element [0] of the tsOut array.
     * 
     * @param ts
     * @param tsOut
     * @return
     */
    public long getIndex(TmfTimestamp ts, TmfTimestamp[] tsOut)
    {
    	TmfTimestamp ts2 = TmfTimestamp.BigCrunch;
    	long index = 0;
    	for( ITmfTrace trace : fTraces )
    	{
    		TmfTraceContext context = trace.seekEvent(ts);
    		index += context.getIndex();
    		TmfTimestamp ts1 = context.getTimestamp();
    		if( ts1 != null && ts2.compareTo( ts1, false ) > 0 )
    			ts2 = ts1;
    	}
   		tsOut[ 0 ] = ts2;
		return index;
    }
    
    /**
     * Returns the timestamp of the event at the requested index.
     * If none, returns null.
     *  
     * @param index
     * @return
     */
    public TmfTimestamp getTimestamp(int index) {
        // TODO: Go over all the traces
    	ITmfTrace trace = fTraces.firstElement();
    	TmfTraceContext context = trace.seekEvent(index);
    	return context.getTimestamp();
    }

    // ========================================================================
    // Operators
    // ========================================================================

    /**
     * Add a trace to the experiment trace set
     * 
     * @param trace
     */
    public void addTrace(ITmfTrace trace) {
		fTraces.add(trace);
		synchronized(this) {
			updateNbEvents();
			updateTimeRange();
		}
    }

    /**
     * Update the total number of events
     */
    protected void updateNbEvents() {
    	int nbEvents = 0;
    	for (ITmfTrace trace : fTraces) {
    		nbEvents += trace.getNbEvents();
    	}
    	fNbEvents = nbEvents;
    }

    /**
     * Update the global time range
     */
    protected void updateTimeRange() {
		TmfTimestamp startTime = /*fTimeRange != null ? fTimeRange.getStartTime() :*/ TmfTimestamp.BigCrunch;
		TmfTimestamp endTime   = /*fTimeRange != null ? fTimeRange.getEndTime()   :*/ TmfTimestamp.BigBang;

		for (ITmfTrace trace : fTraces) {
    		TmfTimestamp traceStartTime = trace.getTimeRange().getStartTime();
    		if (traceStartTime.compareTo(startTime, true) < 0)
    			startTime = traceStartTime;

    		TmfTimestamp traceEndTime = trace.getTimeRange().getEndTime();
    		if (traceEndTime.compareTo(endTime, true) > 0)
    			endTime = traceEndTime;
    	}
		fTimeRange = new TmfTimeRange(startTime, endTime);
    }

    // ========================================================================
    // ITmfRequestHandler
    // ========================================================================

	/* (non-Javadoc)
	 * @see org.eclipse.linuxtools.tmf.request.ITmfRequestHandler#processRequest(org.eclipse.linuxtools.tmf.request.TmfDataRequest, boolean)
	 */
	public void processRequest(TmfDataRequest<TmfEvent> request, boolean waitForCompletion) {

		// Process the request
		processDataRequest(request);
		
		// Wait for completion if needed
    	if (waitForCompletion) {
			request.waitForCompletion();
		}
	}

	/**
	 * Process a data request
	 * 
	 * @param request
	 */
	protected void processDataRequest(final TmfDataRequest<TmfEvent> request) {

		// General request parameters
		final TmfTimestamp endTime;
		final long index;

		// Initialize request params depending on request type
		if (request.getRange() != null) {
			index = Math.max( getIndex(request.getRange().getStartTime()), request.getIndex());
			endTime = request.getRange().getEndTime();
    	} else {
    		index = request.getIndex();
    		endTime = TmfTimestamp.BigCrunch;
    	}

		// Process the request
		Thread thread = new Thread() {

			@Override
			public void run() {

				// Key variables
				ITmfTrace[] traces = new ITmfTrace[0];	// The set of traces
				TmfTraceContext[] contexts;				// The set of trace contexts

				// Extract the general request information
				int blockSize = request.getBlockize();
				int nbRequestedEvents = request.getNbRequestedEvents();
				if (nbRequestedEvents == -1) {
					nbRequestedEvents = Integer.MAX_VALUE;
				}

				// Create the result buffer
				Vector<TmfEvent> events = new Vector<TmfEvent>();
				int nbEvents = 0;

				// Initialize the traces array and position the traces
				// at the first requested event
				traces = fTraces.toArray(traces);
				contexts = new TmfTraceContext[traces.length];
				positionTraces(index, traces, contexts);

				// Get the ordered events
				EventIterator i = getEventIterator( traces, contexts );
				TmfEvent event = i.getNextEvent();
				while (!request.isCancelled() && nbEvents < nbRequestedEvents && event != null
						&& event.getTimestamp().compareTo(endTime, false) < 0)
				{
					events.add(event);
					if (++nbEvents % blockSize == 0) {
						pushData(request, events);
					}
					// Avoid an unnecessary read passed the last event requested
					if (nbEvents < nbRequestedEvents)
						event = i.getNextEvent();
				}

				if (!request.isCancelled() && !request.isFailed()) {
					pushData(request, events);
					request.done();
				}
			}
		};
		thread.start();
	}

	protected EventIterator getEventIterator( ITmfTrace[] traces, TmfTraceContext[] contexts )
	{
		return new EventIterator( traces, contexts );
	}
	
	class EventIterator
	{

		protected TreeSet< NextEvent > nextEvents;
		protected NextEvent nextEvent;

		public EventIterator( ITmfTrace[] traces, TmfTraceContext[] contexts )
		{
			// Setup ordered merging of traces
			nextEvents = new TreeSet< NextEvent >( new OrderNextEventByTimestamp());
			for( int i = 0; i < contexts.length; i++ )
			{
				if( contexts[ i ].getTimestamp() != null )
					nextEvents.add( new NextEvent( i, traces[ i ], contexts[ i ]) );
			}

			nextEvent = null;
		}
		
		protected EventIterator()
		{
		}
		
		public TmfEvent getNextEvent()
		{
			if( nextEvent != null )
			{
				if( nextEvent.context.getTimestamp() != null )
				{
					// re-insert current and fetch new settings only when more than one trace
					// and the old is not the earliest one (not the next one)
					if( !nextEvents.isEmpty() && 
						nextEvents.comparator().compare( nextEvent, nextEvents.first()) >= 0 )
					{
						nextEvents.add( nextEvent );	// re-insert current settings (ordered !!!)
						nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
					}
				} else
					nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
			} else
				nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
			// Fetch next event
			if( nextEvent != null )
			{
				TmfEvent event = nextEvent.trace.getNextEvent( nextEvent.context );
				return event;
			}
			
			return null;
		}
		
	}
	
	static class SearchByIndex implements Comparator<TmfExperimentCheckpoint>
	{

		@Override
		public int compare( TmfExperimentCheckpoint o1, TmfExperimentCheckpoint o2 )
		{
			long r = o1.getIndex() - o2.getIndex();
			if( r < 0 )
				r = -1;
			else if( r > 0 )
				r = +1;
			return (int) r;
		}
		
	}
	
	/**
	 * Given an experiment event index, position the set of traces so a call
	 * to getNextEvent() will retrieve the corresponding event.
	 * 
	 * @param index
	 * @param traces
	 * @param contexts
	 * @param nextEvents
	 */
	protected synchronized void positionTraces(long index, ITmfTrace[] traces, TmfTraceContext[] contexts) {

		// Compute the index page and corresponding index
		int page = Collections.binarySearch( fCheckpoints, new TmfExperimentCheckpoint( index, null, null ), new SearchByIndex() );
		
        // In the very likely case that the checkpoint was not found, bsearch
        // returns its negated would-be location (not an offset...). From that
        // index, we can then position the stream and get the event.
		if( page < 0 ) {
			page = Math.max( 0, -( page + 2 ) );
		}
		
		int current;

		// Retrieve the checkpoint and set the contexts (make copies)
		if( page >= fCheckpoints.size()) {
			// If the page entry doesn't exist (e.g. indexing not completed),
			// set contexts at the the last entry (if it exists)
			page = fCheckpoints.size() - 1;
		}
		if( page >= 0 ) {
			current = (int)fCheckpoints.elementAt( page ).getIndex();
			TmfTraceContext[] saveContexts = fCheckpoints.elementAt(page).getContexts();
			for (int i = 0; i < contexts.length; i++) {
				contexts[i] = new TmfTraceContext(saveContexts[i]);
			}
		} else {
			// Index is empty... position traces at their beginning
			current = 0;
			for (int i = 0; i < contexts.length; i++) {
				contexts[i] = new TmfTraceContext(traces[i].seekLocation(null));
			}
		}

		// Position the traces at the requested index
		/*
		while (current++ < index) {
			getNextEvent(traces, contexts);
		}
		*/
		if( current < index )
		{
			EventIterator i = getEventIterator( traces, contexts );
			while( current++ < index )
				i.getNextEvent();
		}
	}

	/**
	 * Scan the next events from all traces and return the next one
	 * in chronological order.
	 * 
	 * @param traces
	 * @param contexts
	 * @param nextEvents
	 * @return
	 */
	protected TmfEvent getNextEvent(ITmfTrace[] traces, TmfTraceContext[] contexts) {
		// TODO: Consider the time adjustment
		/*
		int trace = 0;
		TmfTimestamp timestamp = contexts[trace].getTimestamp();
		if (timestamp == null) {
			timestamp = TmfTimestamp.BigCrunch;
		}
		for (int i = 1; i < traces.length; i++) {
			if (contexts[i].getTimestamp() != null) {
				TmfTimestamp otherTS = contexts[i].getTimestamp();
				if (otherTS.compareTo(timestamp, true) < 0) {
					trace = i;
					timestamp = otherTS;
				}
			}
		}
		TmfEvent event = traces[trace].getNextEvent(contexts[trace]);
		return event;
		*/
		EventIterator i = getEventIterator( traces, contexts );
		return i.getNextEvent();
	}

	/**
	 * Format the result data and notify the requester.
	 * Note: after handling, the data is *removed*.
	 * 
	 * @param request
	 * @param events
	 */
	protected void pushData(TmfDataRequest<TmfEvent> request, Vector<TmfEvent> events) {
		TmfEvent[] result = new TmfEvent[events.size()];
		events.toArray(result);
		request.setData(result);
		request.handleData();
		events.removeAllElements();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "[TmfExperiment (" + fExperimentId + ")]";
	}

    // ========================================================================
    // Indexing
    // ========================================================================

	/*
	 * The experiment holds the globally ordered events of its set of traces.
	 * It is expected to provide access to each individual event by index i.e.
	 * it must be possible to request the Nth event of the experiment.
	 * 
	 * The purpose of the index is to keep the information needed to rapidly
	 * restore the traces contexts at regular intervals (every INDEX_PAGE_SIZE
	 * event).
	 */

	// The index page size
	protected static final int DEFAULT_INDEX_PAGE_SIZE = 1000;
	protected final int fIndexPageSize;

	// The experiment index
//	private Vector<TmfTraceContext[]> fExperimentIndex = new Vector<TmfTraceContext[]>();
	protected Vector<TmfExperimentCheckpoint> fCheckpoints = new Vector<TmfExperimentCheckpoint>();

	// Indicates that an indexing job is already running
	protected Boolean fIndexing = false;

	// The indexing job
	protected Job job;

	public boolean getIndexing()
	{
		return fIndexing;
	}
	
	/**
	 * indexExperiment
	 * 
	 * Creates the experiment index.
	 */
	public void indexExperiment() {

		synchronized(fIndexing) {
			if (fIndexing) {
				// An indexing job is already running but a new request came
				// in (probably due to a change in the trace set). The index
				// being currently built is therefore already invalid.
				// TODO: Cancel and restart the job
				// TODO: Add support for dynamically adding/removing traces
				return;
			}
			fIndexing = true;
		}

		job = new IndexingJob(fExperimentId);
		job.schedule();
//		try {
//			job.join();
//		} catch (InterruptedException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}
		if (fWaitForIndexCompletion) {
	        ProgressMonitorDialog dialog = new ProgressMonitorDialog(null);
	        try {
				// TODO: Handle cancel!
	            dialog.run(true, true, new IRunnableWithProgress() {
	                public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
	                    monitor.beginTask("Indexing " + fExperimentId, IProgressMonitor.UNKNOWN);
	    				job.join();
	                    monitor.done();
	                }
	            });
	        } catch (InvocationTargetException e) {
	            e.printStackTrace();
	        } catch (InterruptedException e) {
	            e.printStackTrace();
	        }
		}
	}

	private class IndexingJob extends Job {

		public IndexingJob(String name) {
			super(name);
		}

		/* (non-Javadoc)
		 * @see org.eclipse.core.runtime.jobs.Job#run(org.eclipse.core.runtime.IProgressMonitor)
		 */
		@Override
		protected IStatus run(IProgressMonitor monitor) {

			int nbTraces = fTraces.size();
			
			// Minimal check
			if (nbTraces == 0) {
				synchronized( fIndexing )
				{
					fIndexing.notify();
		            fIndexing = false;
				}
				return Status.OK_STATUS;
			}

			monitor.beginTask("Indexing " + fExperimentId, IProgressMonitor.UNKNOWN);

			Vector<TmfExperimentCheckpoint> checkpoints = new Vector<TmfExperimentCheckpoint>();
			
			try
			{
				// get start contexts of all traces
				TmfTraceContext savedContexts[] = new TmfTraceContext[ nbTraces ];
				for( int i = 0; i < nbTraces; i++ )
				{
					ITmfTrace trace = fTraces.get( i );
					TmfTraceContext ctx = trace.seekLocation( null );
					savedContexts[ i ] = new TmfTraceContext( ctx.getLocation(), ctx.getTimestamp(), 0 );
				}
	
				long nextCheckpoint = 0;	// index of next checkpoint
				long nextIndex = 0;			// current index
				TmfTimestamp nextTimestamp = new TmfTimestamp();	// current time stamp
	
				long chkpntToleranceLow = (long)( fIndexPageSize * 0.1 );	// checkpoint will be added if nextIndex >= nextCheckpoint-chkpntToleranceLow
				
				while( true )
				{
					// add checkpoint if we are sufficiently near the expected checkpoint index
					if( nextIndex >= nextCheckpoint - chkpntToleranceLow )
					{
						checkpoints.add( new TmfExperimentCheckpoint( nextIndex, nextTimestamp, savedContexts.clone() ) );
						do
							nextCheckpoint += fIndexPageSize;
						while( nextCheckpoint < nextIndex );
					}

					long skipCount = nextCheckpoint - nextIndex;	// total number of events to skip until next checkpoint
					
					// find number of active traces (those that have data left)
					int nbActiveTraces = 0;
					for( int i = 0; i < nbTraces; i++ )
						if( savedContexts[ i ].getTimestamp() != null )
							nbActiveTraces++;
					if( nbActiveTraces == 0 )
						break;	// done
					
					// special handling for indexing a single trace file
					if( nbActiveTraces == 1 )
					{
						for( int i = 0; i < nbTraces; i++ )
						{
							TmfTraceContext context = savedContexts[ i ];
							if( context.getTimestamp() != null )
							{
								context = new TmfTraceContext( context );
								long skipped = fTraces.get( i ).skipEvents( context, skipCount, null, null, null );
								if( context.getTimestamp() != null )
								{
									Assert.isTrue( skipped >= 0 );	// shall return -1 only when we have been at end, but then we shouldn't be here because timestamp==null at end.
									nextTimestamp = context.getTimestamp();
									nextIndex += skipped;
								}
							}
							savedContexts[ i ] = context;
						}
					}
					else
					{
						// distribute the required number of skipped events evenly over all active traces, and find the one with the smallest timestamp 
						nextTimestamp = TmfTimestamp.BigCrunch;
						int nextTimestampTrace = -1;
						long skipTrace = ( skipCount + nbActiveTraces - 1 ) / nbActiveTraces;
						if( skipTrace <= 0 ) skipTrace = 1;	// just to make sure we're doing anything
						for( int i = 0; i < nbTraces; i++ )
						{
							TmfTraceContext context = savedContexts[ i ];
							if( context.getTimestamp() != null )
							{
								TmfTraceContext nextContext = new TmfTraceContext( context );
								fTraces.get( i ).skipEvents( nextContext, skipTrace, null, null, null );
								if( nextContext.getTimestamp() != null && nextTimestamp.compareTo( nextContext.getTimestamp(), true ) > 0 )
								{
									if( nextContext.getTimestamp().compareTo( context.getTimestamp(), true ) >= 0 )
									{
										nextTimestamp = nextContext.getTimestamp();
										nextTimestampTrace = i;
									}
									else
									{
										// the timestamp *after* skipping is smaller than the timestamp before
										// this indicates a non-monotonic rise and is not allowed
										// we will forcible stop indexing this trace event stream
										savedContexts[ i ].setTimestamp( null );
										// TODO: report some message to the user
									}
								}
							}
						}
						
						if( nextTimestampTrace == -1 )
							break;	// we're done because all traces are exhausted before the next check point is reached

						// check if the target time stamp is different (higher) than the current one 
						if( savedContexts[ nextTimestampTrace ].getTimestamp().compareTo( nextTimestamp, true ) != 0 )
						{
							// given the smallest time stamp, advance all (other) traces to this time stamp, counting the number of skipped events
							// Note: skipEvents using a time stamp will advance the context to a point *after* the given time stamp
							// therefore we need a time stamp immediately *before* the target time stamp
							TmfTimestamp nextTimestamp1 = nextTimestamp.synchronize( -1, (byte)0 );
							for( int i = 0; i < nbTraces; i++ )
							{
								TmfTraceContext context = savedContexts[ i ];
								if( context.getTimestamp() != null )
								{
									context = new TmfTraceContext( context );
									long skipped = fTraces.get( i ).skipEvents( context, skipCount, nextTimestamp1, null, null );
									if( skipped > 0 )
									{
										nextIndex += skipped;
										skipCount -= skipped;
									}
								}
								savedContexts[ i ] = context; 
							}
						}
						else
						{
							// special case is when we don't advance the time stamp, because all events (in the trace with the smallest time)
							// are at the same - the initial - time stamp
							for( int i = 0; i < nbTraces && skipCount > 0; i++ )
							{
								TmfTraceContext context = savedContexts[ i ];
								if( context.getTimestamp() != null )
								{
									context = new TmfTraceContext( context );
									long skipped = fTraces.get( i ).skipEvents( context, skipCount, nextTimestamp, null, null );
									if( skipped > 0 )
									{
										nextIndex += skipped;
										skipCount -= skipped;
									}
								}
								savedContexts[ i ] = context; 
							}
							// we might have advanced all contexts after the target time stamp, so we need to update the nextTimestamp
							// in case this configuration is recorded as checkpoint
							nextTimestamp = TmfTimestamp.BigCrunch;
							for( int i = 0; i < nbTraces; i++ )
							{
								TmfTraceContext context = savedContexts[ i ];
								if( context.getTimestamp() != null && nextTimestamp.compareTo( context.getTimestamp(), true ) > 0 )
									nextTimestamp = context.getTimestamp();
							}
							if( nextTimestamp == TmfTimestamp.BigCrunch )
								break;	// done, no more checkpoints
						}
					}
					
	    			monitor.worked(1);
	                if (monitor.isCanceled()) {
	                    monitor.done();
	                    synchronized( fIndexing )
	        			{
	                        fIndexing.notify();
	                        fIndexing = false;
	        			}
	                	return Status.CANCEL_STATUS;
	                }
				}
			}			
			catch( Throwable t)
			{
				t.printStackTrace();
			}
		
            monitor.done();
            synchronized( fIndexing )
			{
                fCheckpoints = checkpoints;
                fIndexing.notify();
                fIndexing = false;
			}
            return Status.OK_STATUS;
			
			
//			class TraceContextPair implements Comparable<TraceContextPair> {
//				public TraceContextPair( ITmfTrace trace, TmfTraceContext context ) {
//					this.trace = trace;
//					this.context = context;
//				}
//			
//				public ITmfTrace trace;
//				public TmfTraceContext context;
//				
//				@Override
//				public int compareTo( TraceContextPair o )
//				{
//					if( context.getTimestamp() == null )
//						return 1;
//					if( o.context.getTimestamp() == null )
//						return -1;
//					return context.getTimestamp().compareTo( o.context.getTimestamp(), true );
//				}
//			}
//			
//			int nullEvents = 0;
//			
//			ArrayList<TraceContextPair> traceContextList = new ArrayList<TraceContextPair>();
//			TmfTraceContext savedContexts[] = new TmfTraceContext[ nbTraces ];
//    		for (int i = 0; i < nbTraces; i++) {
//    			ITmfTrace trace = fTraces.get( i );
//    			TmfTraceContext ctx = trace.seekLocation(null);
//    			savedContexts[i] =  new TmfTraceContext(ctx.getLocation(), ctx.getTimestamp(), 0);
//    			traceContextList.add( new TraceContextPair( trace, savedContexts[i]));
//    			if (ctx.getTimestamp() == null)
//    				nullEvents++;
//    		}
//			
//    		// Check if there is anything to index
//    		if (nullEvents >= nbTraces) {
//	            fIndexing = false;
//				return Status.OK_STATUS;
//    		}
//
//			TmfTraceContext contexts[] = new TmfTraceContext[nbTraces];
//			int i = 0;
//			for( TmfTraceContext ctx : savedContexts )
//				contexts[i++] = new TmfTraceContext( ctx );
//			indices.add( contexts );
//    		
//    		try {
//				if( traceContextList.size() > 0 ) {	
//					while( true ) {
//						long skipCount = fIndexPageSize;
//						while( skipCount > 0 )
//						{
//							// sort the list of trace/context pairs by the timestamp of the context
//							Collections.sort( traceContextList );
//							
//							TmfTraceContext context = traceContextList.get( 0 ).context;
//							ITmfTrace trace = traceContextList.get( 0 ).trace;
//							TmfTimestamp nextTimestamp = traceContextList.size() > 1 ?
//								nextTimestamp = traceContextList.get( 1 ).context.getTimestamp() : null;
//							
//							long number = trace.skipEvents( context, skipCount, nextTimestamp != null ? nextTimestamp : TmfTimestamp.BigCrunch );
//							if( number == -1 ) {
//								traceContextList.remove( 0 );
//								if( traceContextList.size() == 0 )
//									break;
//								
//							}
//							skipCount -= number;
//						}
//
//						if( traceContextList.size() == 0 )
//							break;
//						
//						contexts = new TmfTraceContext[nbTraces];
//						i = 0;
//						for( TmfTraceContext ctx : savedContexts )
//							contexts[i++] = new TmfTraceContext( ctx );
//						indices.add( contexts );
//						
//		    			monitor.worked(1);
//		                if (monitor.isCanceled()) {
//		                    monitor.done();
//		                	return Status.CANCEL_STATUS;
//		                }
//					}
//				}
//    		}
//			catch( Throwable t)
//			{
//				t.printStackTrace();
//			}
//			
//            monitor.done();
//            fExperimentIndex = indices;
//
////            dumpIndex();
//
//            fIndexing = false;
//            return Status.OK_STATUS;
		}
    }

//	/**
//	 * Dump the experiment index
//	 */
//	private void dumpIndex() {
//		System.out.println("-----");
//		System.out.println("Index of " + fExperimentId);
//		for (int i = 0; i < fExperimentIndex.size(); i++) {
//        	System.out.println("Entry:" + i);
//        	TmfTraceContext[] contexts = fExperimentIndex.get(i);
//        	int nbEvents = 0;
//        	for (int j = 0; j < contexts.length; j++) {
//        		ITmfTrace trace = fTraces.get(j);
//            	TmfTraceContext context = trace.seekLocation(contexts[j].getLocation());
//            	TmfEvent event = fTraces.get(j).getNextEvent(new TmfTraceContext(context));
//            	nbEvents += contexts[j].getIndex();
//            	System.out.println("  ["  + trace.getName() + "]" 
//            			+  " index: "     + contexts[j].getIndex()
//            			+ ", timestamp: " + contexts[j].getTimestamp()
//            			+ ", event: "     + event.getTimestamp());
//            	assert (contexts[j].getTimestamp().compareTo(event.getTimestamp(), false) == 0);
//        	}
//        	assert ((i+1) * fIndexPageSize == nbEvents);
//	
//        }
//	}

    // ========================================================================
    // Signal handlers
    // ========================================================================

//    @TmfSignalHandler
//    public void experimentSelected(TmfExperimentSelectedSignal signal) {
//		fCurrentExperiment = this;
//    	indexExperiment();
//    }

    @TmfSignalHandler
    public void experimentUpdated(TmfExperimentUpdatedSignal signal) {
//    	indexExperiment();
    }

    @TmfSignalHandler
    public void traceUpdated(TmfTraceUpdatedSignal signal) {
    	// TODO: Incremental index update
    	synchronized(this) {
    		updateNbEvents();
    		updateTimeRange();
    	}
		broadcastSignal(new TmfExperimentUpdatedSignal(this, this, signal.getTrace()));
    }

    // ========================================================================
    // Filter and search support
    // ========================================================================

	/* (non-Javadoc)
	 * @see org.eclipse.linuxtools.tmf.request.ITmfRequestHandler#processRequest(org.eclipse.linuxtools.tmf.request.TmfDataRequest, boolean)
	 */
	public void processRequest(TmfFilteringDataRequest<TmfEvent> request, boolean waitForCompletion) {

		// Process the request
		processDataRequest(request);
		
		// Wait for completion if needed
    	if (waitForCompletion) {
			request.waitForCompletion();
		}
	}

	static class NextEvent
	{
		public final int traceIndex;
		public final ITmfTrace trace;
		public final TmfTraceContext context;
		public NextEvent( int traceIndex, ITmfTrace trace, TmfTraceContext context )
		{
			this.traceIndex = traceIndex;
			this.trace = trace;
			this.context = context;
		}
	}
	
	static class OrderNextEventByTimestamp implements Comparator< NextEvent >
	{
		@Override
		public int compare( NextEvent o1, NextEvent o2 )
		{
			int r = o1.context.getTimestamp().compareTo( o2.context.getTimestamp(), true );
			if( r == 0 )
				r = o1.traceIndex - o2.traceIndex;
			return r;
		}
	}

	static class NextEvent2
	{
		final int traceIndex;
		final ITmfTrace trace;
		final TmfTraceContext context;
		TmfTimestamp nextTimestamp;
		int skipped;
		int consumed;
		TmfTimestamp[] skippedTS;
		NextEvent2( int traceIndex, ITmfTrace trace, TmfTraceContext context, TmfTimestamp nextTimestamp, int skipped, TmfTimestamp[] skippedTS )
		{
			this.traceIndex = traceIndex;
			this.trace = trace;
			this.context = context;
			this.nextTimestamp = nextTimestamp;
			this.skipped = skipped;
			this.consumed = 0;
			this.skippedTS = skippedTS;
		}
//		int consumeTo_lin( TmfTimestamp ts, boolean inclusive )
//		{
//			// Linear search
//			int comp = inclusive ? 1 : 0;
//			int start = consumed;
//			while( consumed < skipped && skippedTS[ consumed ].compareTo( ts, false ) < comp )
//				++consumed;
//			return consumed - start;
//		}
		int consumeTo( TmfTimestamp ts, boolean inclusive )
		{
			// Binary search
			int comp = inclusive ? 1 : 0;
			int first = consumed, last = skipped, test;
			while( first != last )
			{
				test = ( first + last ) / 2;
				if( skippedTS[ test ].compareTo( ts, false ) < comp )
					first = test + 1;
				else
					last = test;
			}
			first = last - consumed;
			consumed = last;
			return first;
		}
	}

	static class OrderNextEventByTimestamp2 implements Comparator< NextEvent2 >
	{
		@Override
		public int compare( NextEvent2 o1, NextEvent2 o2 )
		{
			int r = o1.nextTimestamp.compareTo( o2.nextTimestamp, true );
			if( r == 0 )
				r = o1.traceIndex - o2.traceIndex;
			return r;
		}
	}

	/**
	 * Process a data request
	 * 
	 * @param request
	 */
	protected void processDataRequest(final TmfFilteringDataRequest<TmfEvent> request) {

		// Process the request
		Thread thread = new Thread() {

			@Override
			public void run() {

				// General request parameters
				final TmfTimestamp endTime;
				final long index;
		
				// Initialize request params depending on request type
				if (request.getRange() != null) {
					index = Math.max( getIndex(request.getRange().getStartTime()), request.getIndex());
					endTime = request.getRange().getEndTime();
		    	} else {
		    		index = request.getIndex();
		    		endTime = TmfTimestamp.BigCrunch;
		    	}

				// Key variables
				Object filter = request.getFilter();

				// Initialize the traces array and position the traces
				// at the first requested event
				long endIndex = request.getNbTestedItems();
				if( endIndex < 0 || endIndex > Long.MAX_VALUE - index )
					endIndex = Long.MAX_VALUE;
				else
					endIndex += index;

				ExperimentFiltering p = new ExperimentFiltering()
				{
					// General request information
					int blockSize;
					int nbRequestedEvents;

					// Create the result buffer
					Vector<TmfEvent> events = new Vector<TmfEvent>();
					Vector<Integer> indexes = new Vector<Integer>();
					
					@Override
					protected boolean isCancelled()
					{
						return request.isCancelled();
					}
					

					@Override
					protected boolean addMatchingEvent( TmfEvent event, long experimentIndex, long nbMatchingEvents )
					{
						events.add( event );
						indexes.add( (int)experimentIndex );
						if( nbMatchingEvents % blockSize == 0 )
							pushData( request, events, indexes );
						return ( nbMatchingEvents < nbRequestedEvents );	// false = No more matches requested.
					}

					@Override
					public void run( final Object filter, final long startIndex, final long endIndex, final TmfTimestamp endTime )
					{
						// Extract the general request information
						blockSize = request.getBlockize();
						nbRequestedEvents = request.getNbRequestedEvents();
						if (nbRequestedEvents <= -1) {
							nbRequestedEvents = Integer.MAX_VALUE;
						}
						
						super.run( filter, index, endIndex, endTime );
						
						if (!request.isCancelled() && !request.isFailed()) {
							if( events.size() != 0 )
								pushData( request, events, indexes );
							request.done();
						}
					}
				};
				p.run( filter, index, endIndex, endTime );
				
//				System.out.println("found " + nbEvents + " events.");
			}
		};
		thread.start();
	}
	/**
	 * Format the result data and notify the requester.
	 * Note: after handling, the data is *removed*.
	 * 
	 * @param request
	 * @param events
	 */
	protected void pushData(TmfFilteringDataRequest<TmfEvent> request, Vector<TmfEvent> events, Vector<Integer> indexes) {
		TmfEvent[] result = new TmfEvent[events.size()];
		events.toArray(result);
		request.setData(result);
		int[] resultIndexes = new int[indexes.size()];
		int s = indexes.size();
		for( int i = 0; i < s; i++ )
			resultIndexes[ i ] = indexes.get( i );
		request.setDataIndex( resultIndexes );
		request.handleData();
		events.removeAllElements();
		indexes.removeAllElements();
	}

	public abstract class ExperimentFiltering
	{
		
		public void run( final Object filter, final long startIndex, final long endIndex, final TmfTimestamp endTime ) 
		{

			// Key variables
			ITmfTrace[] traces = getTraces();	// The set of traces
			TmfTraceContext[] contexts;				// The set of trace contexts

			int nbEvents = 0;

			// Initialize the traces array and position the traces
			// at the first requested event
			contexts = new TmfTraceContext[ traces.length ];
			positionTraces(startIndex, traces, contexts);
			long curIndex = startIndex;

			// Get the first match result for each trace
			long lookAhead = Math.min( DEFAULT_INDEX_PAGE_SIZE, endIndex - curIndex );
			TreeSet< NextEvent2 > nextEvents = new TreeSet< NextEvent2 >( new OrderNextEventByTimestamp2());
			for (int i = 0; i < traces.length; i++)
				if( contexts[ i ].getTimestamp() != null )	// not at end of trace?
				{
					TmfTimestamp[] skippedTS = new TmfTimestamp[ DEFAULT_INDEX_PAGE_SIZE ];
					int skipped = (int)traces[ i ].skipEvents( contexts[ i ], lookAhead, null, filter, skippedTS );
					TmfTimestamp nextTimestamp = contexts[ i ].getTimestamp();
					if( nextTimestamp == null && skipped != 0 )
						nextTimestamp = TmfTimestamp.BigCrunch;
					if( nextTimestamp != null )
						nextEvents.add(  new NextEvent2( i, traces[ i ], contexts[ i ], nextTimestamp, skipped, skippedTS ));
				}

			if( !nextEvents.isEmpty())
			{
				// Fetch settings of next earliest event
				NextEvent2 nextEvent = nextEvents.pollFirst();
				// Search loop
				while( !isCancelled() && nextEvent != null )
				{
					TmfTimestamp skipTo = nextEvent.context.getTimestamp();
					if( skipTo == null )
					{
						// There is no next match, all traces still in
						// the list do not contribute matching events.
						break;
					}
					if( skipTo.compareTo( endTime, false ) >= 0 )
					{
						// Next match exceeds end time.
						break;
					}
					// Advance all other traces to the point of the earliest
					// event
					for( NextEvent2 otherEvent : nextEvents )
						curIndex += otherEvent.consumeTo( skipTo, otherEvent.traceIndex < nextEvent.traceIndex );
					curIndex += nextEvent.skipped - nextEvent.consumed;
					if( curIndex >= endIndex )
					{
						// We skip block of events, so we might exceed
						// the end index!
						break;
					}
					// The earliest event must not be a matching event, it
					// can also happen that there was no matching event within
					// the look-ahead range.
					lookAhead = Math.min( DEFAULT_INDEX_PAGE_SIZE, endIndex - curIndex ); // new look-ahead range
//					long skipped = nextEvent.trace.skipEvents( nextEvent.context, lookAhead, null, filter );
//					if( skipped == 0 )
//					{
						TmfEvent event = nextEvent.trace.getNextEvent( nextEvent.context );
						if( nextEvent.trace.testEvent( event, filter ) )
						{
							++nbEvents;
							if( !addMatchingEvent( event, curIndex, nbEvents ))
								break; // No more matches requested.
						}
						curIndex++;
						nextEvent.skipped = (int)nextEvent.trace.skipEvents( nextEvent.context, lookAhead, null, filter, nextEvent.skippedTS );
						nextEvent.consumed = 0;
//					} else
//					{
//						nextEvent.skipped = skipped;
//					}

					// test if trace does not make any further contribution
					nextEvent.nextTimestamp = nextEvent.context.getTimestamp();
					if( nextEvent.nextTimestamp == null && nextEvent.skipped > 0 )
						nextEvent.nextTimestamp = TmfTimestamp.BigCrunch;
					if( nextEvent.nextTimestamp != null )
					{
						// re-insert current and fetch new settings only when more than one trace
						// and the old is not the earliest one (not the next one)
						if( nextEvents.isEmpty() || 
							nextEvents.comparator().compare( nextEvent, nextEvents.first()) < 0 )
						{
							continue;	// nextEvents is fine, continue with next loop iteration 
						}
						nextEvents.add( nextEvent );	// re-insert current settings (ordered !!!)
					}
					nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
				}
			}
		}
		
		protected boolean isCancelled()
		{
			return false;
		}
		
		abstract protected boolean addMatchingEvent( TmfEvent event, long experimentIndex, long nbMatchingEvents );
		
	}
	
}